import os
import numpy as np
import pandas as pd
from numpy import dot
from numpy.linalg import norm
from typing import List
import json
from itertools import cycle

import spacy
from spacy import displacy
from spacy.tokens import Doc, Span, Token
from spacy.language import Language

from turcy.pattern import Pattern

def get_root(doc):
    for token in doc:
        if token.dep_  == "ROOT":
            return token

def merge_names(doc):
    with doc.retokenize() as retokenizer:
        for token in doc:
            lefts = [left_c for left_c in token.lefts]
            if len(lefts)>0:
                if lefts[0].tag_ == "NE" and lefts[0].dep_ == "pnc":
                    new_tok = doc[lefts[0].i : token.i+1]
                    retokenizer.merge(new_tok)

    with doc.retokenize() as retokenizer:
        for token in doc:
            rights = [right_c for right_c in token.rights]
            if len(rights)>0:
                if token.tag_ == "NE" and rights[0].tag_ == "NE" and rights[0].dep_ == "nk":
                    right_tok = doc[token.i : rights[0].i+1]
                    retokenizer.merge(right_tok)
    return doc

def merge_noun_chunks(doc):
    with doc.retokenize() as retokenizer:
        for cunk in doc.noun_chunks:
            retokenizer.merge(cunk)
    return doc

def merge_entities(doc):
    pass


class Triple(object):
    def __init__(self, text=None, part=None, pos=None):
        self.text = text
        self.part = part
        self.pos = pos


def isSubtree(mainTree, subTree, triple=None, ll=None) -> List:

    triples = []

    if triple == None:
        triple = []
    if ll == None:
        ll = []

    if (subTree == None):
        if len(triple)>0:
            triples.append(triple)
            return triples
        else:
            return []

    if (mainTree == None):
        return []

    main_tree_lefts = [left_c for left_c in mainTree.lefts]
    main_tree_rights = [rigth_c for rigth_c in mainTree.rights if ( (rigth_c.text != ".")
                                                                    and (rigth_c.text != "(")
                                                                    and (rigth_c.text != ")")
                                                                    and (rigth_c.text != ",")
                                                                    and (rigth_c.text != ":"))]
    sub_tree_lefts = [left_c for left_c in subTree.lefts]
    sub_tree_rights = [rigth_c for rigth_c in subTree.rights if ((rigth_c.tag_ != "$.")
                                                                 and (rigth_c.tag_ != "$(")
                                                                 and (rigth_c.tag_ != "$,"))]

    if len(main_tree_lefts)<=0:
        main_tree_lefts.append(None)
    if len(main_tree_rights)<=0:
        main_tree_rights.append(None)
    if len(sub_tree_lefts)<=0:
        sub_tree_lefts.append(None)
    if len(sub_tree_rights)<=0:
        sub_tree_rights.append(None)

    main_tree = list(zip(main_tree_lefts, cycle(main_tree_rights))
                     if len(main_tree_lefts) > len(main_tree_rights)
                     else zip(cycle(main_tree_lefts), main_tree_rights))
    sub_tree = list(zip(sub_tree_lefts, cycle(sub_tree_rights))
                    if len(sub_tree_lefts) > len(sub_tree_rights)
                    else zip(cycle(sub_tree_lefts), sub_tree_rights))
    ll = list(zip(main_tree, sub_tree))
    for idx, (main, sub) in enumerate(ll):
        left_main_tree, right_main_tree = main
        left_sub_tree, right_sub_tree = sub
        # print(main_tree_left, main_tree_right)

        if ( (mainTree.tag_ == subTree.tag_) and mainTree.tag_ != "$," and mainTree.dep_ == subTree.dep_):
            if mainTree not in triple:
                if subTree.keep:
                    mainTree._.posi = subTree.posi
                    mainTree._.part = subTree.part
                    triple.append(mainTree)
            leftTree = isSubtree(left_main_tree, left_sub_tree, triple[:], ll[:])
            rightTree = isSubtree(right_main_tree, right_sub_tree, triple[:], ll[:])
            triples = isAppendable(leftTree, rightTree, left_sub_tree, right_sub_tree, triples)
            if len(ll)-1 == idx: # or subTree.truth == len(triples[0])
                return triples
        elif (len(triple)>0):
            return []
    triples.extend(isSubtree(left_main_tree, subTree))
    triples.extend(isSubtree(right_main_tree, subTree))
    return triples

def expand_tree_items(tree):
    new_tree = []
    for lis in tree:
        for token in lis:
            if token not in new_tree:
                new_tree.append(token)
    return new_tree

def is_subset(expanded_triple, triples)->bool:
    is_subset = False
    for triple in triples:
        sorted(triple)
        for trip in expanded_triple:
            sorted(trip)
            set1 = set(triple)
            set2 = set(trip)
            is_subset = set2.issubset(set1)
            # return is_subset
    return is_subset

def isAppendable(leftTree, rightTree, leftSubTree, rightSubTree, triples ):
    if ((len(rightTree) > 0 and rightSubTree != None) and (len(leftTree) == 0)):
        if not is_subset(rightTree, triples):
            # right tree has already a part, merge new and old.
            if any(triples):
                right_tree = expand_tree_items(rightTree)
                old_tree = expand_tree_items(triples)
                expanded_triple = [old_tree + [l for l in right_tree if l not in old_tree]]
                return expanded_triple
            else:
                return rightTree
        else:
            return triples
    # 2.2 Found in right only
    elif ((len(leftTree) > 0 and leftSubTree != None) and (len(rightTree) == 0)):
        if not is_subset(leftTree, triples):
            # left tree has already a part, merge new and old.
            if any(triples):
                left_tree = expand_tree_items(rightTree)
                old_tree = expand_tree_items(triples)
                expanded_triple = [old_tree + [l for l in left_tree if l not in old_tree]]
                return expanded_triple
            return leftTree
        else:
            return triples
    # Found some triple, in both, resolve issue
    if len(leftTree) > 0 and len(rightTree) > 0:
        if leftTree == rightTree:
            triples.extend(leftTree)
        elif leftTree[0][0] == rightTree[0][0]:
            left_tree = expand_tree_items(leftTree) # added
            right_tree = expand_tree_items(rightTree)
            expanded_triple = [left_tree + [l for l in right_tree if l not in left_tree]]#added
            if not is_subset(expanded_triple, triples):
                left_tree = expand_tree_items(triples)
                right_tree = expand_tree_items(expanded_triple)
                triples.extend([left_tree + [l for l in right_tree if l not in left_tree]])
                if len(triples) == 2:
                    while len(triples) > 1:
                        del triples[0]
            # triples.extend([leftTree[0] + [l for l in rightTree[0] if l not in leftTree[0]]])
        else:
            left_tree = expand_tree_items(leftTree)
            right_tree = expand_tree_items(rightTree)
            triples.extend([left_tree + [l for l in right_tree if l not in left_tree]])
            if len(triples)==2:
                # right_triple = expand_tree_items(rightTree[1])
                triples.extend([triples[1] + [l for l in triples[0] if l not in triples[1]]])
                while len(triples)>1:
                    del triples[0]
    return triples


#encoding="utf8"
def pattern_loader(path_to_rules="patterns.jsonl") -> List:
    patterns = []
    patternDicts = []
    with open(path_to_rules, encoding="utf8") as f:
        try:
            for line in f:
                data = json.loads(line)
                patternDicts.append(data) # json.loads(f.read())
        except:
            pass
    for patternDict in patternDicts:
        patterns.append(Pattern(**patternDict))
    return patterns

def reassemble(tripleList):
    subj = ""
    subjL = {}
    pred = ""
    predL = {}
    obj = ""
    objL = {}
    for token in tripleList:
        if token._.part == "subj":
            # add it to the right position on subj
            subjL.update({token._.posi: token.text})
        elif token._.part == "pred":
            predL.update({token._.posi: token.text})
        elif token._.part == "obj":
            objL.update({token._.posi: token.text})
        else:
            raise Exception('Triple does not belong to any part (subj,pred,obj).')

    subj = " ".join(str(subjL[elem]) for elem in sorted(subjL))
    pred = " ".join(str(predL[elem]) for elem in sorted(predL))
    obj = " ".join(str(objL[elem]) for elem in sorted(objL))
    if subj != "" and pred != "" and obj != "":
        return (subj, pred, obj)
    else:
        return None

@Language.component('attach_triple2sentence')
def attach_triple2sentence(doc):
    work_dir = os.path.dirname(os.path.realpath(__file__))
    file_dir = os.path.join(work_dir, "patterns.jsonl")
    patternList = pattern_loader(path_to_rules=file_dir)

    for sent in doc.sents:
        for pattern in patternList:
            subTreeMatch = isSubtree(get_root(sent), pattern)
            for tripleList in subTreeMatch:
                if len(tripleList) >= 3 and len(tripleList) == pattern.truth:
                    triples = reassemble(tripleList)
                    if triples != None:
                        sent._.triples.append({"triple": triples, "rule": pattern.name})
    return doc


def add_patterns(nlp):
    nlp.add_pipe(attach_triple2sentence, last=True)
    return nlp


# def extract(user_input):
#     Doc.set_extension('triples', default=[], force=True)
#     Span.set_extension('triples', default=[], force=True)
#
#     nlp = spacy.load("de_core_news_lg")
#     lemma_lookup = nlp.vocab.lookups.get_table("lemma_lookup")
#     lemma_lookup["Amtskollegen"] = "Amtskollege"
#     lemma_lookup["Kanzlerin"] = "Kanzler"

    # "Student Christian Klose kommt aus Deutschland. "  # i1
    # "Der Präsident, Frank-Walter Steinmeier ließ verlauten, dass ... . "  # i2
    # "Barack Obama, einer der Präsidenten, ließ verlauten, dass der Traum von Freiheit tot ist. "  # i3
    # "Angela Merkel ist die Kanzlerin Deutschlands. "  # i4
    # "Der Kanzler Frankreichs heißt Emmanuel Macron. "  # h1
    # "Tobias trifft seinen Bruder Christian Klose. "  # h2
    # "Franz Schneider, dessen Frau Hausfrau ist, muss den Abwasch machen. "  # h3
    # "Die Kanzlerin von Deutschland, ist Frau Merkel."  # h4
    # "In Bayern in Deutschland können Sie vor allem im Frankenland die Natur genießen. "  # p1
    # # "Optimismus ist nur ein Mangel an Erfahrung." p2
    # "Auf Amerikas Präsident Joe Biden wartet schon ein Berg Arbeit. "  # n1
    # "All das gehörte nicht Deutschland, dessen Präsident Christian Wulff war. "  # n2
    # "Angela Merkel schüttelte ihrem Amtskollegen Vladimir Putin die Hand. "  # n3

    # nlp = add_patterns(nlp)
    # doc = nlp(user_input)
    # # doc = attach_triple2sentence(doc, patternList)
    # df = pd.DataFrame(columns=["subj","pred", "obj", "rule"])
    # store = TripleStore()
    # for sent in doc.sents:
    #     for triple in sent._.triples:
    #         (subj, pred, obj) = triple["triple"]
    #         rule = triple["rule"]
    #         df = df.append({'subj': subj, 'pred':pred,'obj':obj,'rule':rule}, ignore_index=True)
    #         #st.write(subj, type(subj))
    #         #st.write(pred, type(pred))
    #         #st.write(obj, type(obj))
    #         # st.write(f"{subj}, {pred}, {obj}, {rule}")
    #         store.add_triple(subj, pred, obj, "")
    #         #st.write(triple)
    #         #tripleList.append(triple)
    # return df, store


# Drei Bruchpiloten in Paris (Originaltitel: La Grande vadrouille), in Deutschland auch unter dem Titel Die große Sause bekannt, ist eine französische Filmkomödie, bei der Gérard Oury Regie führte und die 1966 in die Kinos kam. The Drop – Bargeld ist ein US-amerikanisches Kriminalfilm-Drama des belgischen Regisseurs Michaël R. Roskam aus dem Jahr 2014. Das Drehbuch stammt von Autor Dennis Lehane, der dazu seine eigene Kurzgeschichte Animal Rescue verarbeitete. Im Jahr 1987 zählte Köttweinsdorf 117 Einwohner. Die EN 60601-2-27 mit dem Titel „Medizinische elektrische Geräte – Teil 2-27: Besondere Festlegungen für die Sicherheit von Elektrokardiographie-Überwachungsgeräten“ ist Teil der Normenreihe EN 60601. Herausgeber der DIN-Norm DIN EN 60601-2-27 ist das Deutsche Institut für Normung. Der HC Lev Poprad war ein 2010 gegründeter Eishockeyklub mit Sitz in der slowakischen Stadt Poprad. Quers ist eine Gemeinde im französischen Département Haute-Saône in der Region Bourgogne-Franche-Comté. Bacoor (offiziell: Municipality of Bacoor; Filipino: Bayan ng Bakoor) ist eine philippinische Stadtgemeinde in der Provinz Cavite. Der Lagonda 2.6 Litre war ein Oberklassefahrzeug, das der britische Automobilhersteller Aston Martin von 1948 bis 1953 unter dem Markennamen Lagonda anbot. Poręba Żegoty (örtlich Porymba Zegoty) ist ein Dorf mit einem Schulzenamt der Gemeinde Alwernia im Powiat Chrzanowski der Woiwodschaft Kleinpolen in Polen. Becklingen liegt etwa 7 km nördlich von Bergen an der Bundesstraße 3 und hat 362 Einwohner (Stand: 31. Juli 2015). Die Gmina Abramów ist eine Landgemeinde im Powiat Lubartowski der Woiwodschaft Lublin in Polen. Der Pegaso Z-102 ist ein Sportwagen, der von 1951 bis 1958 in Spanien gebaut wurde. Parbayse ist eine französische Gemeinde mit Einwohnern (Stand ) im Département Pyrénées-Atlantiques in der Region Nouvelle-Aquitaine (vor 2016: Aquitanien). Der Wahlbezirk Böhmen 102 war ein Wahlkreis für die Wahlen zum Abgeordnetenhaus im österreichischen Kronland Böhmen. Laihia (finnisch; schwedisch Laihela) ist eine Gemeinde in der westfinnischen Landschaft Österbotten. Tropico 4 ist eine von Haemimont Games entwickelte und vom Publisher Kalypso Media vertriebene Wirtschaftssimulation aus dem Jahr 2011. Das Spiel ist der vierte Teil der Wirtschaftsimulations-Reihe Tropico und der direkte Nachfolger von Tropico  3. Der Spieler übernimmt wie in den vorangegangenen Teilen die Rolle von „El Presidente“, dem Diktator der fiktiven Bananenrepublik Tropico. Das Hempstead County ist ein     County im Bundesstaat Arkansas. Der Unsichtbare kehrt zurück (Originaltitel: The Invisible Man Returns) ist ein US-amerikanischer Horror-/Science-Fiction-Film des Regisseurs Joe May aus dem Jahr 1940. Die Universal-Produktion wurde als Fortsetzung zu Der Unsichtbare (1933) gedreht und basiert lose auf einem Roman von H. G. Wells. Schigansk (; , Edjigeen) ist ein Dorf (selo) in der Republik Sacha (Jakutien) in Russland mit Einwohnern (Stand ). Neuküstrinchen ist ein Ortsteil der Gemeinde Oderaue im Landkreis Märkisch-Oderland in Brandenburg. Linsdorf ist eine französische Gemeinde mit Einwohnern (Stand ) im Département Haut-Rhin in der Region Grand Est (bis 2015 Elsass). Cartigny-l’Épinay ist eine französische Gemeinde mit Einwohnern (Stand ) im Département Calvados in der Region Normandie. Filipiñana ist ein philippinisch-britischer Spielfilm unter der Regie von Rafael Manuel aus dem Jahr 2020. Ausbeutung und Ausgrenzung in den hierarchischen Strukturen der philippinischen Gesellschaft werden am Beispiel eines Golfclubs gezeigt. Das Schlangennest (engl. Titel: Whacking Day) ist die zwanzigste Folge der vierten Staffel der Serie Die Simpsons. Tłustoręby (deutsch Kirchberg) ist ein Dorf in der Gmina Niemodlin, im Powiat Opolski, der Woiwodschaft Oppeln im Südwesten von Polen. Die Schweizerische Gesellschaft für Organisation und Management (SGO) ist ein Verein mit Sitz in Zürich, Schweiz. Kleinkarlbach ist eine Ortsgemeinde im pfälzischen Landkreis Bad Dürkheim.
# Das Hempstead County ist ein County im US-Bundesstaat Arkansas.
# Die Gmina Abramów ist eine Landgemeinde im Powiat Lubartowski.

# Drei Bruchpiloten in Paris (Originaltitel: La Grande vadrouille), in Deutschland auch unter dem Titel Die große Sause bekannt, ist eine französische Filmkomödie, bei der Gérard Oury Regie führte und die 1966 in die Kinos kam.


import os
import sys
import re
import json
import pprint

# import legalSentencizer
from turcy.utils import save_jsonl
import spacy
from turcy.pattern import Pattern
import string

def set_part(pattern, token, trip_d, trl_part_len=list):
    idx = None
    elems = []
    subj = trip_d[0]
    pred = trip_d[1]
    obj = trip_d[2]

    if token.text in [tok.text for tok in subj] and trl_part_len[0]>0:
        pattern.part = "subj"
        for idx, tok in enumerate([tok for tok in subj]):
            if token.text == tok.text:
                pattern.posi = subj[idx].i
                trl_part_len[0] -= 1
    elif token.text in [tok.text for tok in pred] and trl_part_len[1]>0:
        pattern.part = "pred"
        for idx, tok in enumerate([tok for tok in pred]):
            if token.text == tok.text:
                pattern.posi = pred[idx].i
                trl_part_len[1] -= 1
    elif token.text in [tok.text for tok in obj] and trl_part_len[2]>0:
        pattern.part = "obj"
        for idx, tok in enumerate([tok for tok in obj]):
            if token.text == tok.text:
                pattern.posi = obj[idx].i
                trl_part_len[2] -= 1

def loopTree(pattern, rootPattern):
    update_completness(pattern, rootPattern)
    for left in pattern.lefts:
        loopTree(left, rootPattern)
    for right in pattern.rights:
        loopTree(right, rootPattern)
    if rootPattern.subj and rootPattern.pred and rootPattern.obj:
        rootPattern.complete = True
        return 1
    else:
        return 0

def update_completness(pattern, rootPattern):
    if pattern.part == "subj":
        rootPattern.subj = True
        rootPattern.truth+=1
    elif pattern.part == "pred":
        rootPattern.pred = True
        rootPattern.truth += 1
    elif pattern.part == "obj":
        rootPattern.obj = True
        rootPattern.truth += 1

def build(token, patternTree=None, path="", triple=None, first=False, trip_d=None, count_key=None, trl_part_len=None):
    pattern = Pattern(tag=token.tag_, dep=token.dep_, keep=False, name=count_key)
    if first != True:
        patternTree[path].append(pattern)
    else:
        patternTree = pattern
    if token.text in triple:
        set_part(pattern, token, trip_d, trl_part_len)
        pattern.keep = True
        triple.remove(token.text)
    elif any(token.text in word for word in triple):
        part = [part for part in triple if token.text in part]
        part_idx = triple.index(part[0])
        words = part[0].split()
        for word in words:
            if word == token.text:
                pattern.keep = True
                set_part(pattern, token, trip_d, trl_part_len)
                triple[part_idx] = triple[part_idx].replace(token.text, '').strip()
                break
        for idx, trip in enumerate(triple):
            if trip.isspace() or all(i in string.punctuation for i in trip):
                del triple[idx]
        if len(triple)==0:
            return patternTree
    lefts = [token for token in token.lefts]
    rights = [token for token in token.rights]
    for left in lefts:
        patternTree = build(left, pattern, triple=triple, path="lefts", first=False, trip_d=trip_d, trl_part_len=trl_part_len)
    for right in rights:
        patternTree = build(right, pattern, triple=triple, path="rights", first=False, trip_d=trip_d, trl_part_len=trl_part_len)
    return patternTree


def doc_length(doc):
    length = len([token for token in doc])
    return length

nlp = spacy.load('de_core_news_lg')

def find(sent, triple, key=""):
    count = 0
    # legalSentencizer.add_to_pipe(nlp)
    count_key = key + "_" + str(count)
    count+=1
    subj = triple[0][2]
    pred = triple[1][2]
    obj  = triple[2][2]
    trip = [subj, pred, obj]
    subj_doc = nlp(triple[0][2])
    pred_doc = nlp(triple[1][2])
    obj_doc  = nlp(triple[2][2])
    trip_d = [subj_doc, pred_doc, obj_doc]
    doc = nlp(sent)
    for token in doc:
        if token.dep_ == "ROOT":
            root_token = token
            break
    trl_part_len = [doc_length(subj_doc), doc_length(pred_doc), doc_length(obj_doc)] #trl_part_len
    pattern = build(root_token, triple=trip, first=True, trip_d=trip_d, count_key=count_key, trl_part_len=trl_part_len)
    # check if it's a complete pattern
    loopTree(pattern, pattern)
    return pattern

def add(pattern):
    save_jsonl(pattern, "patterns.jsonl")


import os
import re
import json
import pandas as pd
import urllib.request
import pickle
import itertools
from itertools import islice

def save_data(data, file_name):
    path = os.path.dirname(os.path.realpath(__file__))
    with open(f'{path}/{file_name}.pickle', 'wb') as handle:
        pickle.dump(data, handle, protocol=pickle.HIGHEST_PROTOCOL)

# @st.cache(suppress_st_warning=True)
def load_data(file_name=None, path=None):
    # if ((file_name == None) and (path==None)):
        # dataset = load_dataset("wikipedia", '20200501.de', split='train')
        # titles = dataset["title"]
        # texts = dataset["text"]
        # glitter = zip(titles, texts)
        #  data = dict(glitter)
        # return data, titles, texts
    if ((file_name != None) and (path !=None)):
        with open(f'{file_name}.pickle', 'rb') as handle:
            data = pickle.load(handle)
            # return dataset
    elif ((file_name == None) and (path != None)):
        work_dir = os.path.dirname(os.path.realpath(__file__))
        file_dir = os.path.join(work_dir, path)
        with open(file_dir, 'rb') as handle:
            data = pickle.load(handle)
            # return dataset
    return data

#########
#########
#########
def save_jsonl(file, path_filename):
    """
    Write list of objects to a JSON lines file.
    """
    work_dir = os.path.dirname(os.path.realpath(__file__))
    file_dir = os.path.join(work_dir, path_filename)
    with open(f"{file_dir}", 'w', encoding="utf8") as outfile:
        for entry in file.items():
            json.dump(entry, outfile)
            outfile.write('\n')
    print('Wrote {} records to {}'.format(len(file), path_filename))

def save_json(data_dict, file_name=None, path=None):
    if ((file_name != None) and (path == None)):
        path = os.path.dirname(os.path.realpath(__file__))
        with open(f"{path}/{file_name}.json", 'w', encoding="utf8") as f:
            json.dumps(data_dict, indent=4)
    elif ((file_name == None) and (path != None)):
        work_dir = os.path.dirname(os.path.realpath(__file__))
        file_dir = os.path.join(work_dir, path)
        with open(f"{file_dir}", 'w', encoding="utf8") as f:
            json.dump(data_dict, f, indent=4)

def load_json(path=None, file_name=None):
    if ((file_name != None) and (path == None)):
        path = os.path.dirname(os.path.realpath(__file__))
        with open(f"{path}/{file_name}.json", encoding="utf8") as f:
                train_data = json.loads(f.read())
    elif ((file_name == None) and (path != None)):
        work_dir = os.path.dirname(os.path.realpath(__file__))
        file_dir = os.path.join(work_dir, path)
        with open(file_dir, encoding="utf8") as f:
            train_data = json.loads(f.read())
    return train_data

def split_dict_equally(input_dict, chunks=2):
    """
    Splits dict by keys. Returns a list of dictionaries.
    """
    return_list = [dict() for idx in range(chunks)]
    idx = 0
    for k,v in input_dict.items():
        return_list[idx][k] = v
        if idx < chunks-1:
            idx += 1
        else:
            idx = 0
    return return_list

def chunks(data, SIZE=10000):
    it = iter(data)
    for i in range(0, len(data), SIZE):
        yield {k:data[k] for k in islice(it, SIZE)}



import os
import logging
import typing
import spacy
from spacy.tokens import Doc, Span, Token
from turcy.tree_dep_pattern import attach_triple2sentence
from spacy.language import Language

logging.basicConfig(level=logging.INFO)

Doc.set_extension('triples', default=[], force=True)
Span.set_extension('triples', default=[], force=True)
Token.set_extension('part', default=None, force=True)
Token.set_extension('posi', default=None, force=True)

def add_to_pipe(nlp):
    nlp.add_pipe("attach_triple2sentence", last=True)


from typing import List
import json


class Pattern(dict):
    def __init__(self, tag_=None,
                 dep_=None,
                 lefts=None,
                 rights=None,
                 name=None,
                 keep=True,
                 appendOrder="inOrder",
                 part=None,
                 posi=None,
                 subj=None,
                 pred=None,
                 obj=None,
                 complete=None,
                 truth=0,
                 **entries):
        object.__init__(self)
        self.tag_ = tag_
        self.dep_ = dep_
        self.lefts = lefts if lefts is not None else []
        self.rights = rights if rights is not None else []
        self.name = name
        self.keep = keep
        self.appendOrder = appendOrder  # Pre-order #https://medium.com/odscjournal/binary-search-tree-implementation-in-python-5f8a50341eaf
        self.part = part
        self.posi = posi
        self.subj = subj
        self.pred = pred
        self.obj = obj
        self.complete = complete
        self.truth = truth
        # self.__dict__.update(entries)

    def __getattr__(self, name):
        if name in self:
            return self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __setattr__(self, name, value):
        if name == "lefts" or name == "rights":
            self[name] = []
            for val in value:
                self[name].append(Pattern(**val))
        else:
            self[name] = value

    def __delattr__(self, name):
        if name in self:
            del self[name]
        else:
            raise AttributeError("No such attribute: " + name)

    def __str__(self):
        return f"{{'tag':'{self.tag_}','dep':'{self.dep_}'}}"

    def __repr__(self):
        return f"{{'tag':'{self.tag_}','dep':'{self.dep_}'}}"



import setuptools

# python -m build

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="turcy", # Replace with your own username
    version="0.0.25",
    author="Christian Klose",
    author_email="chris.klose@gmx.net",
    description="A package for German Open Informtion Extraction",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ChrisChross/turCy",
    packages=setuptools.find_packages(),
    data_files=[('', ['turcy/patterns.jsonl'])],
    include_package_data=True,
    keywords="openie turcy information extraction spacy",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)